% Name  :  Robust Logistic Regression
% Author:  Jakramate Bootkrajang
% Last update: 22 August 2019
% Input :  
%          w       = given initial model's parameters
%          g       = initial label flipping probability matrix
%          x       = Design matrix where row represents a sample
%          y       = Target value
%          regFunc = type of regularisation: lasso (l1) or l2 
%          options.estG  = estimate the flipping matrix
%          options.verbose = display negative log-likelihood
% Output:  
%          w     = fitted model
%          g     = estimated gamma matrix
%          llh   = negative log-likelihood
% Note  :
%       1. The model uses {0,1} class representation internally. 
%       w = (eye(size(x,2))-x'*inv(x*x'+eye(size(x,1)))*x)*x'*y/2;  w(1) = 1;
%  |-----|-----------|
%  |     |  0   y^  1|
%  ------|-----------|
%  |y  0 | g00   g01 |
%  |   1 | g10   g11 |
%  |-----|-----------|
%=====================================================================================            

function [w g llh] = rlr(w, g, x, y, regFunc, options)

% do some input preprocessing
% check if the problem is binary problem
if (size(unique(y),1) ~= 2)
    error('Trying to solve multiclass problem');
end

% check if bias term is added i.e. first 
% column must be unique and equal to 1
if (sum(x(:,1)) ~= size(x,1))
    disp('Bias terms might have not been added');
end

if ~isfield(options,'sn')
    options.sn = 1e-8;
end

% ========================= BEGIN ESTIMATING PARAMETERS =========================== 
for l=1:50   

    t  = x * w; 
                
    % update gamma matrix
    if options.estG
        for gl=1:3
            s0  = (g(1,1) * (1 ./ ((1 ./ exp(-t))+1))) + (g(2,1) ./ (1+exp(-t)));
            s1  = (g(1,2) * (1 ./ ((1 ./ exp(-t))+1))) + (g(2,2) ./ (1+exp(-t)));
            
            % avoid numerical problem
            s0(s0==0) = eps;
            s1(s1==0) = eps;
            
            % update the gammas
            g00 = g(1,1) * sum(((1 - y) ./ (s0)) .* (1 ./ ((1 ./ exp(-t))+1)),1);
            g01 = g(1,2) * sum((y ./ s1) .* (1 ./ ((1 ./ exp(-t))+1)),1);
            
            % (g00 + g01) is the Lagrangian
            g(1,1) = g00 / (g00 + g01); 
            g(1,2) = 1 - g(1,1);
            
            g10 = g(2,1) * sum(((1-y) ./ s0) ./(1+exp(-t)),1);
            g11 = g(2,2) * sum(((y) ./ (s1)) ./(1+exp(-t)),1);
            
            % (g10 + g11) is the Lagrangian
            g(2,1) = g10 / (g10 + g11); 
            g(2,2) = 1 - g(2,1);
        end
    else
        %% EM E-STEP
        P1 =  (1 ./ (1+exp(-t))) .* (repmat(g(2,2),size(t),1).^y) ...
            .* (repmat(g(2,1),size(t),1).^(1-y));
        P2 =  (1-(1 ./ (1+exp(-t)))) .* (repmat(g(1,2),size(t),1).^y) ...
            .* (repmat(g(1,1),size(t),1).^(1-y));
        bigP = P1./(P1+P2);
        
        %% EM M-STEP
        g(2,2) = (sum(bigP.*y)) / sum(bigP);
        g(2,1) = 1 - g(2,2);
        g(1,1) = (sum((1-bigP).*y))/sum(1-bigP);
        g(1,2) = 1 - g(1,1);
    end
        
    switch func2str(regFunc)
        case 'lasso'
            lambda    = 1./(sqrt(w.^2 + options.sn));    
            lambda(1) = 0;            
        case 'l2'                       
            lambda    = ones(length(w),1) * (length(w)/2 +1)/(sum(w(2:end).^2)/2 + 2); 
            lambda(1) = 0; 
        otherwise
            disp('Invalid regularisation function. Now using lasso');
            regFunc = 'lasso';
    end
    if l==1
        lambda = lambda * 0;
    end
    
    opts.maxIter    = 10;
    opts.Display    = false;    
    [w, fv, v1, v2] = minFunc(@gradw, w, opts, g, x, y, regFunc, lambda, options.sn);       
    
    llh(l)  = fv(end);  % save log-likelihood    
            
    if options.verbose
        disp(llh(l));
    end
end


% =========================== END ESTIMATING PARAMETER ============================

% deal with class identity problem
s = 1;
for loop = 1:2
    [un, i] = max(g(s,:));
    if (g(s,i) > g(i,i))
        g = circshift(g,1);
        w = -w;
    else
        s = s + 1;
    end
end
